/** \file Functors.h */

#ifndef AMDIS_FUNCTORS_H
#define AMDIS_FUNCTORS_H

#include "AbstractFunction.h"

namespace AMDiS {

template<typename T>
struct Id : public AbstractFunction<T,T>
{
  Id() : AbstractFunction<T,T>(1) {}
  T operator()(const T &v) const { return v; }
};

template<typename T1,typename T2>
struct Const : public AbstractFunction<T1,T2>
{
  Const(T1 val_) : AbstractFunction<T1,T2>(0), val(val_) {}
  T1 operator()(const T2 &v) const { return val; }
private:
  T1 val;
};

template<typename T=double>
struct Factor : public AbstractFunction<T,T>
{
  Factor(double fac_) : AbstractFunction<T,T>(1), fac(fac_) {}
  T operator()(const T& x) const { return fac*x; };
private:
  double fac;
};

template<typename T=double>
struct Add : public BinaryAbstractFunction<T,T,T>
{
  Add() : BinaryAbstractFunction<T,T,T>(1) {}
  T operator()(const T &v1, const T &v2) const { return v1+v2; }
};

template<typename T=double>
struct AddFactor : public BinaryAbstractFunction<T,T,T>
{
  AddFactor(double factor_ = 1.0) : BinaryAbstractFunction<T,T,T>(1), factor(factor_) {}
  T operator()(const T &v1, const T &v2) const { return v1 + factor*v2; }
private:
  double factor;
};

template<typename T=double>
struct Subtract : public BinaryAbstractFunction<T,T,T>
{
  Subtract() : BinaryAbstractFunction<T,T,T>(1) {}
  T operator()(const T &v1, const T &v2) const { return v1-v2; }
};

template<typename T=double>
struct AddScal : public AbstractFunction<T,T>
{
  AddScal(T scal_) : AbstractFunction<T,T>(1), scal(scal_) {}
  T operator()(const T &v) const { return v+scal; }
private:
  T scal;
};

template<typename T>
struct Mult : public BinaryAbstractFunction<T,T,T>
{
  Mult() : BinaryAbstractFunction<T,T,T>(2) {}
  T operator()(const T &v1, const T &v2) const { return v1*v2; }
};

template<typename T1, typename T2, typename T3>
struct Mult2 : public BinaryAbstractFunction<T1,T2,T3>
{
  Mult2() : BinaryAbstractFunction<T1,T2,T3>(2) {}
  T1 operator()(const T2 &v1, const T3 &v2) const { return v1*v2; }
};

template<typename T>
struct MultScal : public BinaryAbstractFunction<T,T,T>
{
  MultScal(T scal_) : BinaryAbstractFunction<T,T,T>(2), scal(scal_) {}
  T operator()(const T &v1, const T &v2) const { return v1*v2*scal; }
private:
  T scal;
};

template<typename T>
struct Max : public BinaryAbstractFunction<T,T,T>
{
  Max() : BinaryAbstractFunction<T,T,T>(1) {}
  T operator()(const T &v1, const T &v2) const { return std::max(v1,v2); }
};

template<typename T>
struct Min : public BinaryAbstractFunction<T,T,T>
{
  Min() : BinaryAbstractFunction<T,T,T>(1) {}
  T operator()(const T &v1, const T &v2) const { return std::min(v1,v2); }
};

template<typename T>
struct Diff : public BinaryAbstractFunction<T,T,T>
{
  Diff() : BinaryAbstractFunction<T,T,T>(1) {}
  T operator()(const T &v1, const T &v2) const { return abs(v1-v2); }
};

template<typename T=double>
struct Abs : public AbstractFunction<T,T>
{
  Abs() : AbstractFunction<T,T>(1) {}
  T operator()(const T &v) const { return abs(v); }
};

template<typename T=double>
struct Signum : public AbstractFunction<T,T>
{
  Signum() : AbstractFunction<T,T>(0) {}
  T operator()(const T &v) const { return (v>0.0?1.0:(v<0.0?-1.0:0.0)); }
};

template<typename T=double>
struct Sqr : public AbstractFunction<T,T>
{
  Sqr() : AbstractFunction<T, T>(2) {}
  T operator()(const T &v) const { return sqr(v); }
};

template<typename T=double>
struct Sqrt : public AbstractFunction<T,T>
{
  T operator()(const T &v) const { return sqrt(v); }
};

namespace detail {
  template<int p>
  struct Pow
  {
    template<typename T>
    static T eval(const T& v) { return v*Pow<p-1>::eval(v); }
  };
  
  template<>
  struct Pow<1> {
    template<typename T>
    static T eval(const T& v) { return v; }
  };
  
  template<>
  struct Pow<0> {
    template<typename T>
    static T eval(const T& v) { return 1.0; }
  };
}

template<int p, typename T=double>
struct Pow : public AbstractFunction<T,T>
{
  Pow(double factor_=1.0) : AbstractFunction<T,T>(p), factor(factor_) {}
  T operator()(const T &v) const { return factor * detail::Pow<p>::eval(v); }
private:
  double factor;
};

template<typename T1, typename T2 = ProductType<T1, T1> >
struct Norm2 : public AbstractFunction<T1, T2>
{
  Norm2() : AbstractFunction<T1, T2>(2) {}
  T1 operator()(const T2 &v) const { return sqrt(v*v); }
};

template<typename T1, typename T2>
struct Norm2Sqr : public AbstractFunction<T1, T2>
{
  Norm2Sqr() : AbstractFunction<T1, T2>(2) {}
  T1 operator()(const T2 &v) const { return v*v; }
};

template<typename T>
struct Norm2_comp2 : public BinaryAbstractFunction<T,T,T>
{
  Norm2_comp2() : BinaryAbstractFunction<T,T,T>(2) {}
  T operator()(const T &v1, const T &v2) const { return sqrt(sqr(v1)+sqr(v2)); }
};

template<typename T>
struct Norm2Sqr_comp2 : public BinaryAbstractFunction<T,T,T>
{
  Norm2Sqr_comp2() : BinaryAbstractFunction<T,T,T>(2) {}
  T operator()(const T &v1, const T &v2) const { return sqr(v1)+sqr(v2); }
};

template<typename T>
struct Norm2_comp3 : public TertiaryAbstractFunction<T,T,T,T>
{
  Norm2_comp3() : TertiaryAbstractFunction<T,T,T,T>(2) {}
  T operator()(const T &v1, const T &v2, const T &v3) const { return sqrt(sqr(v1)+sqr(v2)+sqr(v3)); }
};

template<typename T>
struct Norm2Sqr_comp3 : public TertiaryAbstractFunction<T,T,T,T>
{
  Norm2Sqr_comp3() : TertiaryAbstractFunction<T,T,T,T>(2) {}
  T operator()(const T &v1, const T &v2, const T &v3) const { return sqr(v1)+sqr(v2)+sqr(v3); }
};

template<typename T>
struct L1Diff : public BinaryAbstractFunction<T,T,T>
{
  T operator()(const T &v1, const T &v2) const { return abs(v1-v2); }
};

template<typename TOut, typename T=TOut>
struct L2Diff : public BinaryAbstractFunction<TOut,T,T>
{
  TOut operator()(const T &v1, const T &v2) const { return Norm2<TOut, T>()(v1-v2); }
};

template<typename T>
struct Vec1WorldVec : public AbstractFunction<WorldVector<T>,T>
{
  WorldVector<T> operator()(const T &v0) const {
    WorldVector<T> result;
    result[0]=v0;
    return result;
  }
};
template<typename T>
struct Vec2WorldVec : public BinaryAbstractFunction<WorldVector<T>,T,T>
{
  WorldVector<T> operator()(const T &v0, const T &v1) const {
    WorldVector<T> result;
    result[0]=v0; result[1]=v1;
    return result;
  }
};
template<typename T>
struct Vec3WorldVec : public TertiaryAbstractFunction<WorldVector<T>,T,T,T>
{
  WorldVector<T> operator()(const T &v0, const T &v1, const T &v2) const {
    WorldVector<T> result;
    result[0]=v0; result[1]=v1; result[2]=v2;
    return result;
  }
};
template<int c, typename T=double>
struct Component : public AbstractFunction<T, WorldVector<T> >
{
  Component() : AbstractFunction<T, WorldVector<T> >(1) {}
  T operator()(const WorldVector<T> &x) const {
    return x[c];
  }
};

struct FadeOut : public TertiaryAbstractFunction<double, double, double ,double>
{
  double operator()(const double &v, const double &dist, const double &mean) const {
    return dist*mean+(1.0-dist)*v;
  }
};
}

#endif // AMDIS_FUNCTORS_H

