//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include <stdio.h>
#include <string>
#include <fstream>
#include <sstream>
#include <cmath>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/lexical_cast.hpp>

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#include <mpi.h>
#endif

#include "VtkWriter.h"
#include "DataCollector.h"
#include "DOFVector.h"
#include "SurfaceRegion_ED.h"
#include "ElementRegion_ED.h"
#include "AdaptInfo.h"

namespace AMDiS { 

  using namespace std;

  int VtkWriter::writeFile(string name)
  {
    FUNCNAME("VtkWriter::writeFile()");
    
    boost::iostreams::filtering_ostream file;
    switch (compress) {
    case GZIP:
      file.push(boost::iostreams::gzip_compressor());
      name.append(".gz");
      break;
    case BZIP2:
      file.push(boost::iostreams::bzip2_compressor());
      name.append(".bz2");
      break;
    default:
      break;
    }

    {
      ofstream swapfile(name.c_str(), ios::out | ios::trunc);
      TEST_EXIT(swapfile.is_open())
	("Cannot open file %s for writing!\n", name.c_str());
      swapfile.close();
    }

    file.push(boost::iostreams::file_descriptor_sink(name, ios::trunc));
    writeFileToStream(file);

    return 0;
  }


  void VtkWriter::writeParallelFile(string name, int nRanks,
				    string fnPrefix, string fnPostfix,
				    vector<string> &componentNames)
  {
    FUNCNAME("VtkWriter::writeParallelFile()");

    using boost::lexical_cast;

    boost::iostreams::filtering_ostream file;
    {
      ofstream swapfile(name.c_str(), ios::out | ios::trunc);
      TEST_EXIT(swapfile.is_open())
	("Cannot open file %s for writing!\n", name.c_str());
      swapfile.close();
    }
    file.push(boost::iostreams::file_descriptor_sink(name, ios::trunc)); 

    file << "<?xml version=\"1.0\"?>\n";
    file << "<VTKFile type=\"PUnstructuredGrid\">\n";
    file << "  <PUnstructuredGrid GhostLevel=\"0\">\n";
    file << "    <PPoints>\n"
	 << "      <PDataArray type=\"Float32\" NumberOfComponents=\"3\" format=\"ascii\"/>\n"
	 << "    </PPoints>\n";
    file << "    <PCells>\n"
	 << "      <PDataArray type=\"Int32\" Name=\"offsets\"/>\n"
	 << "      <PDataArray type=\"UInt8\" Name=\"types\"/>\n"
	 << "      <PDataArray type=\"Int32\" Name=\"connectivity\"/>\n"
	 << "    </PCells>\n";
    file << "    <PPointData>\n";

    for (unsigned int i = 0; i < componentNames.size(); i++)
      file << "      <PDataArray type=\"Float32\" Name=\"" 
	   << componentNames[i] 
	   << "\" format=\"ascii\"/>\n";

    file << "    </PPointData>\n";

    for (int i = 0; i < nRanks; i++) {
      string pname(fnPrefix + "-p" + lexical_cast<string>(i) + "-" + fnPostfix);
      boost::filesystem::path filepath(pname);
      file << "    <Piece Source=\"" 
	   << boost::filesystem::basename(filepath)
	   << boost::filesystem::extension(filepath) << "\"/>\n";      
    }

    file << "  </PUnstructuredGrid>\n";
    file << "</VTKFile>\n";
  }


  int VtkWriter::updateAnimationFile(AdaptInfo *adaptInfo,
				     string valueFilename,
				     vector< pair<double, string> > *paraViewAnimationFrames,
				     string animationFilename)
  {
    FUNCNAME("VtkWriter::updateAnimationFile()");

    size_t found = valueFilename.find_last_of("/\\");
    paraViewAnimationFrames->push_back(make_pair(adaptInfo->getTime(),
						 valueFilename.substr(found + 1)));

    boost::iostreams::filtering_ostream file;
    {
      ofstream swapfile(animationFilename.c_str(), 
			     ios::out | ios::trunc);
      TEST_EXIT(swapfile.is_open())
	("Cannot open file %s for writing!\n", animationFilename.c_str());
      swapfile.close();
    }
    file.push(boost::iostreams::file_descriptor_sink(animationFilename, 
						     ios::trunc)); 

    file << "<?xml version=\"1.0\"?>\n";
    file << "<VTKFile type=\"Collection\" version=\"0.1\" >"  << "\n";
    file << "<Collection>\n";

    for (vector<pair<double, string> >::iterator it = paraViewAnimationFrames->begin();
	 it < paraViewAnimationFrames->end(); ++it) {
      file << "<DataSet timestep=\"" << it->first
	   << "\" part=\"0\" file=\"" << it->second << "\"/>\n";
    }

    file << "</Collection>\n";
    file << "</VTKFile>\n";

    return 0;
  }


  void VtkWriter::writeFile(DOFVector<double> *values, 
			    string filename,
			    bool writeParallel)
  {
    DataCollector<> dc(values->getFeSpace(), values);
    vector<DataCollector<>*> dcList(0);
    dcList.push_back(&dc);
    writeFile(dcList, filename, writeParallel);
  }


  void VtkWriter::writeFile(vector<DOFVector<double>* > &values,
			    string filename,
			    bool writeParallel)
  {
    vector<DataCollector<>*> dcList(0);
    for (unsigned int i = 0; i < values.size(); i++)
      dcList.push_back(new DataCollector<>(values[i]->getFeSpace(), values[i]));	
    writeFile(dcList, filename, writeParallel);
    for (unsigned int i = 0; i < values.size(); i++)
      delete dcList[i];	
  }


  void VtkWriter::writeFile(WorldVector<DOFVector<double>* > &values,
			    string filename,
			    bool writeParallel)
  {
    vector<DataCollector<>*> dcList(0);
    for (int i = 0; i < values.getSize(); i++)
      dcList.push_back(new DataCollector<>(values[i]->getFeSpace(), values[i]));	
    writeFile(dcList, filename, writeParallel);
    for (int i = 0; i < values.getSize(); i++)
      delete dcList[i];	
  }


  void VtkWriter::writeFile(DOFVector<WorldVector<double> > *values,
			    string filename,
			    bool writeParallel)
  {
    WorldVector<DOFVector<double>*> valuesWV;
    for (int i =0 ; i < valuesWV.getSize(); i++)
      valuesWV[i] = new DOFVector<double>(values->getFeSpace(), "values["+boost::lexical_cast<std::string>(i)+"]");
    transform(values, &valuesWV);
    writeFile(valuesWV, filename, writeParallel);
    for (int i = 0; i < valuesWV.getSize(); i++)
      delete valuesWV[i];
  }

  
  void VtkWriter::writeFile(SystemVector *values, 
			    string filename,
			    bool writeParallel)
  {
    vector<DataCollector<>*> dcList(0);
    for (int i = 0; i < values->getSize(); i++)
      dcList.push_back(new DataCollector<>(values->getDOFVector(i)->getFeSpace(), 
					 values->getDOFVector(i)));    
    writeFile(dcList, filename, writeParallel);
    for (unsigned i = 0; i < dcList.size(); i++)
      delete dcList[i];    
  }
  
  
  void VtkWriter::writeFile(vector<DataCollector<>*> &dcList,
			    string filename,
			    bool writeParallel)
  {
    VtkWriter writer(&dcList);
    
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    if (writeParallel) {
      using boost::lexical_cast;
      
      int sPos = filename.find(".vtu");
      TEST_EXIT(sPos >= 0)("Failed to find file postfix!\n");
      string name = filename.substr(0, sPos);
      
      if (MPI::COMM_WORLD.Get_rank() == 0) {
	vector<string> componentNames;
	for (unsigned int i = 0; i < dcList.size(); i++)
	  componentNames.push_back(dcList[i]->getValues()->getName());

	writer.writeParallelFile(name + ".pvtu", 
				 MPI::COMM_WORLD.Get_size(), 
				 name, ".vtu",
				 componentNames);
      }
      
      filename = name + "-p" + lexical_cast<string>(MPI::COMM_WORLD.Get_rank()) + "-.vtu";
    }
#endif
    writer.writeFile(filename);
  }
}
