// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file PetscSolverNSCH.h */

#include "parallel/PetscSolverGlobalMatrix.h"

namespace AMDiS {

  using namespace std;

  struct CahnHilliardData2 {
    KSP kspMassCH, kspLaplaceCH, kspVelocity, kspLaplace, kspMass;
    Mat matMassCH, matMinusDeltaK, matGrad, matDiv, matConDif, matSchur, velocityMat;
    PetscSolverGlobalMatrix *globalMatrixSolver;
    double *eps, *delta;
    MPI::Intracomm *mpiCommGlobal;
  };
  
      
 class PetscSolverNSCH : public PetscSolverGlobalMatrix
  {
  public:
    /// Creator class
    
        class IdFct : public AbstractFunction<double, double>
    {
    public:
      IdFct() 
	: AbstractFunction<double, double>(0)
      {}
      
      double operator()(const double& x) const 
      {
	return x;
      }
    };

    class MultConstFct : public AbstractFunction<double, double>
    {
    public:
      MultConstFct(double c) 
	: AbstractFunction<double, double>(1),
	  mConst(c)
      {}
      
      double operator()(const double& x) const 
      {
	return mConst * x;
      }
    private:
      double mConst;
    };
    
     class LinearInterpolation : public AbstractFunction<double, double>
    {
    public:
      LinearInterpolation(double c1, double c2, double factor=1.0) 
	: AbstractFunction<double, double>(0)	    
      { a = (c1-c2)/2.0*factor; b = (c1+c2)/2.0*factor; cmin=std::min(c1,c2)*factor; cmax=std::max(c1,c2)*factor;}
      
      double operator()(const double& x) const 
      {	double result = b+a*x;
        if (result<cmin) result = cmin;
	if (result>cmax) result = cmax;
	return result;
      }
    private:
      double a,b,cmin,cmax;
    };
    
   
    class LinearInterpolation2 : public BinaryAbstractFunction<double, double, double>
    {
    public:
      LinearInterpolation2(double c1, double c2, double factor=1.0) 
	: BinaryAbstractFunction<double, double, double>(0)
	  { a = (c1-c2)/2.0*factor; b = (c1+c2)/2.0*factor; cmin=std::min(c1,c2)*factor; cmax=std::max(c1,c2)*factor;}
      
      double operator()(const double& u, const double& x) const 
      {
	double result = b+a*x;
        if (result<cmin) result = cmin;
	if (result>cmax) result = cmax;
	return result * u;	
      }
    private:
      double a,b,cmin,cmax;
    };
    
    
    class Multiplier3 : public BinaryAbstractFunction<double, double, double>
    {
    public:
      Multiplier3() 
	: BinaryAbstractFunction<double, double, double >(0) 
      {}
      
      double operator()(const double& phi, const double& phase) const 
      {
   	return  phase * phi;
      }
    };


    class EinsMinus : public AbstractFunction<double, double>
    {
    public:
      EinsMinus(double d)   
	: AbstractFunction<double, double>(0),
	  c(d)
      {}
      
      double operator()(const double& x) const 
      {
	return c * std::max(1.0-x,0.000001);
      }
    private:
      double c;
    };

    struct Multiplication : public BinaryAbstractFunction<double, double, double>{
      double operator()(const double &v1, const double &v2) const{
	return v2 / v1;
      }
    };
    

    
    
    class Creator : public OEMSolverCreator
    {
    public:
      virtual ~Creator() {}

      /// Returns a new PetscSolver object.
      OEMSolver* create() 
      { 
	return new PetscSolverNSCH(this->name); 
      }
    };

    PetscSolverNSCH(string name);

    void solvePetscMatrix(SystemVector &vec, AdaptInfo *adaptInfo);
  
    void setChData(double *epsPtr, double *deltaPtr)
    {
      eps = epsPtr;
      delta = deltaPtr;
    }
    
    
    void setStokesData(double *invTauPtr, SystemVector *vec, double *nu1_=NULL, double *nu2_=NULL, double *rho1_=NULL, double *rho2_=NULL)
    {
      invTau = invTauPtr;
      solution = vec;
      nu1=nu1_;
      nu2=nu2_;
      rho1=rho1_;
      rho2=rho2_;      
      

    }
    
   
    
    void setPhase(DOFVector<double> *d, double eP3=0)
    {
      phase = d;
      epsPhase3 = eP3;
    }

  protected:
    void initSolver(KSP &ksp);

    void initPreconditioner(PC pc);

    void exitPreconditioner(PC pc);

  private:
    int pressureComponent, chComponent, velocityComponent;

    bool pressureNullSpace;

    /// If true, old solution is used for initial guess in solver phase.
    bool useOldInitialGuess;

    /// 0: approximate solve   1: direct solver
    int velocitySolutionMode;

    /// 0: approximate solve   1: direct solver
    int massSolutionMode;

    /// 0: approximate solve   1: direct solver
    int laplaceSolutionMode;
    
    /// 0: approximate solve   1: direct solver
    int  regularizeLaplace;

    PetscSolver *massMatrixSolverCH, *laplaceMatrixSolverCH, *deltaKMatrixSolver;

    CahnHilliardData2 matShellContext;

    double *eps, *delta;
       
    LinearInterpolation *densityFunctionTau;
    LinearInterpolation *viscosityFunction; 
    LinearInterpolation2 *densityFunction; 
    
    PetscSolver *massMatrixSolver, *laplaceMatrixSolver, *conDifMatrixSolver;

    double *nu, *invTau, *nu1,*nu2,*rho1,*rho2;
    
    double epsPhase3;

    SystemVector* solution;

    DOFVector<double>* phase;
    
    IdFct idFct;
    
  };

}

