/** \file MeshIndicator.h */

#ifndef MESH_INDICATOR_H
#define MESH_INDICATOR_H

#include "AMDiS.h"

using namespace AMDiS;

namespace tools {

  struct MeshIndicator
  {
    MeshIndicator(WorldVector<double> dimension_) : boundary_index(0), dimension(dimension_) { }

    MeshIndicator(Mesh* mesh) : boundary_index(0)
    {
      dimension = Helpers::getMeshDimension(mesh);
    }

    virtual bool operator()(const WorldVector<double>& x) const
    {
      return true;
    }

    virtual bool operator()(const ElInfo* elInfo) const
    {
      bool result = true;
      for (int i = 0; i < elInfo->getMesh()->getDim()+1; i++) {
	result = result && operator()(elInfo->getCoord(i));
	if (!result)
	  break;
      }
      return true;
    }

    void set_boundary_index(int boundary_index_)
    {
      boundary_index = boundary_index_;
    }

  protected:

    bool on_boundary()
    {
      return boundary_index != 0;
    }

  protected:

    int boundary_index;
    WorldVector<double> dimension;
  };


  template<typename T>
  T integrate(const DOFVector<T> &vec, MeshIndicator *meshIndicator)
  {
    FUNCNAME("integrate(MeshIndicator)");

    TEST_EXIT(meshIndicator)("No MeshIndicator defined!\n");

    int deg = 2 * vec.getFeSpace()->getBasisFcts()->getDegree();
    Quadrature* quad =
      Quadrature::provideQuadrature(vec.getFeSpace()->getMesh()->getDim(), deg);
    FastQuadrature *fastQuad =
      FastQuadrature::provideFastQuadrature(vec.getFeSpace()->getBasisFcts(), *quad, INIT_PHI);

    mtl::dense_vector<T> qp(fastQuad->getNumPoints());

    T value; nullify(value);

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_DET;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(vec.getFeSpace()->getMesh(), -1, traverseFlag);
    while (elInfo) {
      if ((*meshIndicator)(elInfo)) {
	vec.getVecAtQPs(elInfo, quad, fastQuad, qp);
	T tmp; nullify(tmp);
	for (int iq = 0; iq < fastQuad->getNumPoints(); iq++) {
	  tmp += fastQuad->getWeight(iq) * qp[iq];
	}
	value += tmp * elInfo->getDet();
      }
      elInfo = stack.traverseNext(elInfo);
    }

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    mpi::globalAdd(value);
  #endif

    return value;
  }


  template<typename TOut>
  TOut integrate_Coords(const FiniteElemSpace* feSpace,
		    AbstractFunction<TOut, WorldVector<double> > *fct, MeshIndicator *meshIndicator)
  {
    FUNCNAME("integrate_Coords()");

    TEST_EXIT(fct)("No function defined!\n");
    TEST_EXIT(meshIndicator)("No MeshIndicator defined!\n");

    int deg = std::max(fct->getDegree(), feSpace->getBasisFcts()->getDegree());
    Quadrature* quad =
      Quadrature::provideQuadrature(feSpace->getMesh()->getDim(), deg);
    FastQuadrature *fastQuad =
      FastQuadrature::provideFastQuadrature(feSpace->getBasisFcts(), *quad, INIT_PHI);

    WorldVector<double> coords;

    TOut value; nullify(value);
    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_DET;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1, traverseFlag);
    while (elInfo) {
      if ((*meshIndicator)(elInfo)) {
	TOut tmp; nullify(tmp);
	for (int iq = 0; iq < fastQuad->getNumPoints(); iq++) {
	  elInfo->coordToWorld(fastQuad->getLambda(iq), coords);
	  tmp += fastQuad->getWeight(iq) * (*fct)(coords);
	}

	value += tmp * elInfo->getDet();
      }
      elInfo = stack.traverseNext(elInfo);
    }

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    mpi::globalAdd(value);
  #endif

    return value;
  }

  template<typename TOut, typename T>
  TOut integrate_Vec(const DOFVector<T> &vec, AbstractFunction<TOut, T> *fct, MeshIndicator *meshIndicator)
  {
    FUNCNAME("integrate(MeshIndicator)");

    TEST_EXIT(fct)("No function defined!\n");
    TEST_EXIT(meshIndicator)("No MeshIndicator defined!\n");

    int deg = std::max(fct->getDegree(),
			2 * vec.getFeSpace()->getBasisFcts()->getDegree());

    Quadrature* quad =
      Quadrature::provideQuadrature(vec.getFeSpace()->getMesh()->getDim(), deg);
    FastQuadrature *fastQuad =
      FastQuadrature::provideFastQuadrature(vec.getFeSpace()->getBasisFcts(), *quad, INIT_PHI);

    mtl::dense_vector<T> qp(fastQuad->getNumPoints());

    TOut value; nullify(value);

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_DET;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(vec.getFeSpace()->getMesh(), -1, traverseFlag);
    while (elInfo) {
      if ((*meshIndicator)(elInfo)) {
	vec.getVecAtQPs(elInfo, quad, fastQuad, qp);
	TOut tmp; nullify(tmp);
	for (int iq = 0; iq < fastQuad->getNumPoints(); iq++) {
	  tmp += fastQuad->getWeight(iq) * (*fct)(qp[iq]);
	}
	value += tmp * elInfo->getDet();
      }
      elInfo = stack.traverseNext(elInfo);
    }

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    mpi::globalAdd(value);
  #endif

    return value;
  }


  template<typename TOut, typename T1, typename T2>
  TOut integrate_Vec2(const DOFVector<T1> &vec1,
		    const DOFVector<T2> &vec2,
		    BinaryAbstractFunction<TOut, T1, T2> *fct, MeshIndicator *meshIndicator)
  {
    FUNCNAME("intDualmesh()");

    TEST_EXIT(fct)("No function defined!\n");
    TEST_EXIT(meshIndicator)("No MeshIndicator defined!\n");

    int deg = std::max(fct->getDegree(),
			2 * vec1.getFeSpace()->getBasisFcts()->getDegree());
    Quadrature* quad =
      Quadrature::provideQuadrature(vec1.getFeSpace()->getMesh()->getDim(), deg);
    FastQuadrature *fastQuad =
      FastQuadrature::provideFastQuadrature(vec1.getFeSpace()->getBasisFcts(), *quad, INIT_PHI);

    mtl::dense_vector<T1> qp1(fastQuad->getNumPoints());
    mtl::dense_vector<T2> qp2(fastQuad->getNumPoints());

    TOut value; nullify(value);
    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_DET;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(vec1.getFeSpace()->getMesh(), -1, traverseFlag);
    while (elInfo) {
      if ((*meshIndicator)(elInfo)) {
	vec1.getVecAtQPs(elInfo, quad, fastQuad, qp1);
	vec2.getVecAtQPs(elInfo, quad, fastQuad, qp2);

	TOut tmp; nullify(tmp);
	for (int iq = 0; iq < fastQuad->getNumPoints(); iq++)
	  tmp += fastQuad->getWeight(iq) * (*fct)(qp1[iq], qp2[iq]);
	value += tmp * elInfo->getDet();
      }
      elInfo = stack.traverseNext(elInfo);
    }

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    mpi::globalAdd(value);
  #endif

    return value;
  }


  template<typename T1, typename T2>
  typename ProductType<T1,T2>::type integrate_VecTimesCoords(const DOFVector<T1> &vec,
		    AbstractFunction<T2, WorldVector<double> > *fct, MeshIndicator *meshIndicator)
  {
    FUNCNAME("integrate_VecTimesCoords()");

    TEST_EXIT(fct)("No function defined!\n");
    TEST_EXIT(meshIndicator)("No MeshIndicator defined!\n");

    typedef typename ProductType<T1,T2>::type TOut;

    const FiniteElemSpace *feSpace = vec.getFeSpace();
    Mesh *mesh = feSpace->getMesh();

    int deg = std::max(fct->getDegree(), 2 * feSpace->getBasisFcts()->getDegree());
    Quadrature* quad = Quadrature::provideQuadrature(mesh->getDim(), deg);
    FastQuadrature *fastQuad =
      FastQuadrature::provideFastQuadrature(feSpace->getBasisFcts(), *quad, INIT_PHI);

    mtl::dense_vector<T1> qp(fastQuad->getNumPoints());
    WorldVector<double> coords;

    TOut value; nullify(value);
    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_DET;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while (elInfo) {
      if ((*meshIndicator)(elInfo)) {
	vec.getVecAtQPs(elInfo, quad, fastQuad, qp);

	TOut tmp; nullify(tmp);
	for (int iq = 0; iq < fastQuad->getNumPoints(); iq++) {
	  elInfo->coordToWorld(fastQuad->getLambda(iq), coords);
	  tmp += fastQuad->getWeight(iq) * (qp[iq] * (*fct)(coords));
	}

	value += tmp * elInfo->getDet();
      }
      elInfo = stack.traverseNext(elInfo);
    }

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    mpi::globalAdd(value);
  #endif

    return value;
  }


  template<typename TOut, typename T>
  TOut integrate_VecAndCoords(const DOFVector<T> &vec,
			    BinaryAbstractFunction<TOut, T, WorldVector<double> > *fct, MeshIndicator *meshIndicator)
  {
    FUNCNAME("integrate_VecAndCoords()");

    TEST_EXIT(fct)("No function defined!\n");
    TEST_EXIT(meshIndicator)("No MeshIndicator defined!\n");

    int deg = std::max(fct->getDegree(),
			2 * vec.getFeSpace()->getBasisFcts()->getDegree());
    Quadrature* quad =
      Quadrature::provideQuadrature(vec.getFeSpace()->getMesh()->getDim(), deg);
    FastQuadrature *fastQuad =
      FastQuadrature::provideFastQuadrature(vec.getFeSpace()->getBasisFcts(), *quad, INIT_PHI);

    mtl::dense_vector<T> qp(fastQuad->getNumPoints());
    WorldVector<double> coordsAtQP;

    TOut value; nullify(value);

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_DET;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(vec.getFeSpace()->getMesh(), -1, traverseFlag);
    while (elInfo) {
      if ((*meshIndicator)(elInfo)) {
	vec.getVecAtQPs(elInfo, quad, fastQuad, qp);
	TOut tmp; nullify(tmp);
	for (int iq = 0; iq < fastQuad->getNumPoints(); iq++) {
	  elInfo->coordToWorld(fastQuad->getLambda(iq), coordsAtQP);
	  tmp += fastQuad->getWeight(iq) * (*fct)(qp[iq], coordsAtQP);
	}
	value += tmp * elInfo->getDet();
      }
      elInfo = stack.traverseNext(elInfo);
    }

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    mpi::globalAdd(value);
  #endif

    return value;
  }


  template<typename TOut, typename T>
  TOut integrateGeneral(const std::vector<DOFVector<T>*> &vecs,
		       AbstractFunction<TOut, std::vector<T> > *fct, MeshIndicator *meshIndicator)
  {
    FUNCNAME("integrateGeneral()");

    TEST_EXIT(fct)("No function defined!\n");
    TEST_EXIT(vecs.size()>0)("No DOFVectors provided!\n");

    int deg = std::max(fct->getDegree(),
		       2 * vecs[0]->getFeSpace()->getBasisFcts()->getDegree());
    Quadrature* quad =
      Quadrature::provideQuadrature(vecs[0]->getFeSpace()->getMesh()->getDim(), deg);
    FastQuadrature *fastQuad =
      FastQuadrature::provideFastQuadrature(vecs[0]->getFeSpace()->getBasisFcts(), *quad, INIT_PHI);

    std::vector<mtl::dense_vector<T> > qp(vecs.size());
    std::vector<T> qp_local(vecs.size());
    for (size_t i = 0; i < vecs.size(); i++)
      qp[i].change_dim(fastQuad->getNumPoints());

    TOut value; nullify(value);
    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_DET;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(vecs[0]->getFeSpace()->getMesh(), -1, traverseFlag);
    while (elInfo) {
      if ((*meshIndicator)(elInfo)) {
	for (size_t i = 0; i < vecs.size(); i++)
	  vecs[i]->getVecAtQPs(elInfo, quad, fastQuad, qp[i]);

	TOut tmp; nullify(tmp);
	for (int iq = 0; iq < fastQuad->getNumPoints(); iq++) {
	  for (size_t i = 0; i < vecs.size(); i++)
	    qp_local[i] = qp[i][iq];
	  tmp += fastQuad->getWeight(iq) * (*fct)(qp_local);
	}
	value += tmp * elInfo->getDet();
      }
      elInfo = stack.traverseNext(elInfo);
    }

    #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    mpi::globalAdd(value);
    #endif

    return value;
  }
  
} // namespace tools

#endif // MESH_INDICATOR_H
