/** \file POperators_SOT.h */

#ifndef P_OPERATORS_SOT_H
#define P_OPERATORS_SOT_H

#include "AMDiS.h"

using namespace std;
using namespace AMDiS;

/// < factor*phase*grad(u) , grad(psi) >
class Phase_SOT : public SecondOrderTerm 
{
public:
/// Constructor.
Phase_SOT(DOFVectorBase<double>* phaseDV, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);

/// Implements SecondOrderTerm::getLALt().
inline void getLALt(const ElInfo *elInfo, std::vector<mtl::dense2D<double> > &LALt) const;    

/// Implenetation of SecondOrderTerm::eval().
inline void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

/// Implenetation of SecondOrderTerm::weakEval().
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
		std::vector<WorldVector<double> > &result);

virtual double f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
double fac;
};

/* -------------------------------------------------------------- */

class Mobility_SOT : public SecondOrderTerm
{
public:
/// Constructor.
Mobility_SOT(DOFVectorBase<double> *rhoDV_, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 

void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* rhoDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> rho;
double factor, eps, rho0, density, mobilityDerivative;

};

/* -------------------------------------------------------------- */

class MobilityCH_SOT : public SecondOrderTerm
{
public:
/// Constructor.
MobilityCH_SOT(DOFVectorBase<double> *chDV_, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 

void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
double factor, delta;

};

/* -------------------------------------------------------------- */

class MobilityCHPhase_SOT : public Phase_SOT
{
public:
/// Constructor.
MobilityCHPhase_SOT(DOFVectorBase<double>* phaseDV, DOFVectorBase<double> *chDV_, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
double delta;

};

/* -------------------------------------------------------------- */

class MobilityCH2_SOT : public SecondOrderTerm
{
public:
/// Constructor.
MobilityCH2_SOT(DOFVectorBase<double> *chDV_, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 

void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
double factor, delta;

};

/* -------------------------------------------------------------- */

class ViscosityPhase_SOT : public Phase_SOT
{
public:
/// Constructor.
ViscosityPhase_SOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *chDV_, double v1_=1.0, double v2_=1.0, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
double v1,v2;

};

/* -------------------------------------------------------------- */

class ViscosityPhase2_SOT : public Phase_SOT
{
public:
/// Constructor.
ViscosityPhase2_SOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *chDV_, double v1_=1.0, double v2_=1.0, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
double v1,v2;

};

/* -------------------------------------------------------------- */

class Viscosity_SOT : public SecondOrderTerm
{
public:
/// Constructor.
Viscosity_SOT(DOFVectorBase<double> *chDV_, double v1_=1.0, double v2_=1.0, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 

void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
double v1,v2;
double factor;

};

/* -------------------------------------------------------------- */

class Viscosity2_SOT : public SecondOrderTerm
{
public:
/// Constructor.
Viscosity2_SOT(DOFVectorBase<double> *chDV_, double v1_=1.0, double v2_=1.0, double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 

void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);

inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
double v1,v2;
double factor;

};

/* -------------------------------------------------------------- */

/// < factor*phase*A*grad(u) , grad(psi) > ... 'I' kann ueberladen werden
class MatrixPhase_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixPhase_SOT(DOFVectorBase<double>* phaseDV, int comp, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
inline void lDlt(const DimVec<WorldVector<double> >& Lambda, 
            const WorldVector<double>& vec,
            mtl::dense2D<double>& LALt,
            double factor) const;
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 

void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);

void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);

inline WorldVector<double> f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
int comp;
double fac;
bool symmetric;
};

/* -------------------------------------------------------------- */

/// < factor*()_ij*grad(u) , grad(psi) >
class MatrixIJ_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixIJ_SOT(int row, int col, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL) {}
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL) {}
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 
void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);
void setFactor(const double fac_) { fac=fac_; }

protected:
int row,col;
double fac;
bool symmetric;
};

/* -------------------------------------------------------------- */

/// < factor*phase*A*grad(u) , grad(psi) > ... 'I' kann ueberladen werden
class MatrixIJPhase_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixIJPhase_SOT(DOFVectorBase<double>* phaseDV, int row, int col, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 
void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
int row,col;
double fac;
bool symmetric;
};

/* -------------------------------------------------------------- */

/// < factor*phase*A*grad(u) , grad(psi) > ... 'I' kann ueberladen werden
class MatrixIJViscosityPhase_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixIJViscosityPhase_SOT(DOFVectorBase<double>* phaseDV, DOFVectorBase<double>* chDV, int row, int col, double v1_, double v2_, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 
void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);
void setFactor(const double fac_) { fac=fac_; }
inline double f(const int iq) const;

protected:
DOFVectorBase<double> *phaseDV;
DOFVectorBase<double>* chDV;

mtl::dense_vector<double> ch;
mtl::dense_vector<double> phase;
int row,col;
double v1,v2;
double fac;
bool symmetric;
};

/* -------------------------------------------------------------- */

/// < factor*A*grad(u) , grad(psi) > ... 'I' kann ueberladen werden
class MatrixIJViscosity_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixIJViscosity_SOT(DOFVectorBase<double>* chDV, int row, int col, double v1_, double v2_, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 
void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);
void setFactor(const double fac_) { fac=fac_; }
inline double f(const int iq) const;

protected:
DOFVectorBase<double>* chDV;

mtl::dense_vector<double> ch;
int row,col;
double v1,v2;
double fac;
bool symmetric;
};

/* -------------------------------------------------------------- */

/// < factor*phase*A*grad(u) , grad(psi) > ... 'I' kann ueberladen werden
class MatrixIJViscosityPhase2_SOT : public SecondOrderTerm
{
public:
/// Constructor
MatrixIJViscosityPhase2_SOT(DOFVectorBase<double>* phaseDV, DOFVectorBase<double>* chDV, int row, int col, double v1_, double v2_, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad=NULL);
void getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const; 
void eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor);
void weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result);
void setFactor(const double fac_) { fac=fac_; }
inline double f(const int iq) const;

protected:
DOFVectorBase<double> *phaseDV;
DOFVectorBase<double>* chDV;

mtl::dense_vector<double> ch;
mtl::dense_vector<double> phase;
int row,col;
double v1,v2;
double fac;
bool symmetric;
};

#endif // P_OPERATORS_SOT_H
