/** \file SingularDirichletBC.h */

#ifndef SINGULAR_DIRICHLET_BC_H
#define SINGULAR_DIRICHLET_BC_H

#include "AMDiS.h"

namespace AMDiS {

/**
 * data structure that holds DOF-Index and double value for singular dirichlet boundary condition.
 **/
struct SingularDirichletBC {
  SingularDirichletBC(int i_, int j_, DegreeOfFreedom idx_, double value_) :
    row(i_), col(j_), idx(idx_), value(value_) {}

  int row, col;
  DegreeOfFreedom idx;
  double value;
};

/**
 * data structure that holds all periodic associations.
 **/
struct ManualPeriodicBC {
  ManualPeriodicBC(size_t row_, std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &indices_) :
    row(row_), indices(indices_) {}

  size_t row;
  std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > indices;
};

namespace details {
    
  /**
   * traverses the mesh and stores all DegreeOfFreedom, that lie on the boundary with index 'nr'
   * and have coordinates that give true in the 'meshIndicator', into a vector of 'indices'.
   **/
  inline void getBoundaryIndices(const FiniteElemSpace* feSpace,
			  BoundaryType nr,
			  AbstractFunction<bool, WorldVector<double> >* meshIndicator,
			  std::set<DegreeOfFreedom> &indices)
  {
    Mesh* mesh = feSpace->getMesh();
    int dim = mesh->getDim();
    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int numBasFcts = basFcts->getNumber();
    std::vector<DegreeOfFreedom> localIndices(numBasFcts);

    indices.clear();

    WorldVector<double> coord;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1,
      Mesh::CALL_LEAF_EL | Mesh::FILL_BOUND | Mesh::FILL_COORDS);
    while (elInfo) {
      for (int face = 0; face < dim + 1; face++) {
	if (elInfo->getBoundary(face) == nr) {
	  basFcts->getLocalIndices(elInfo->getElement(), feSpace->getAdmin(), localIndices);

	  for (int i = 0; i < numBasFcts; i++) {
	    elInfo->coordToWorld(*(basFcts->getCoords(i)), coord);
	    if ((*meshIndicator)(coord))
	      indices.insert(localIndices[i]);
	  }
	  break;
	}
      }

      elInfo = stack.traverseNext(elInfo);
    }
  }

  /**
   * traverses the mesh and stores all DegreeOfFreedom, that have coordinates that give true
   * in the 'meshIndicator', into a vector of 'indices'.
   **/
  inline void getBoundaryIndices(const FiniteElemSpace* feSpace,
			  AbstractFunction<bool, WorldVector<double> >* meshIndicator,
			  std::set<DegreeOfFreedom> &indices)
  {
    Mesh* mesh = feSpace->getMesh();
    int dim = mesh->getDim();
    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int numBasFcts = basFcts->getNumber();
    std::vector<DegreeOfFreedom> localIndices(numBasFcts);

    indices.clear();

    WorldVector<double> coord;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1,
      Mesh::CALL_LEAF_EL | Mesh::FILL_BOUND | Mesh::FILL_COORDS);
    while (elInfo) {
      for (int face = 0; face < dim + 1; face++) {
	if (elInfo->getBoundary(face) != 0) {
	  basFcts->getLocalIndices(elInfo->getElement(), feSpace->getAdmin(), localIndices);

	  for (int i = 0; i < numBasFcts; i++) {
	    elInfo->coordToWorld(*(basFcts->getCoords(i)), coord);
	    if ((*meshIndicator)(coord))
	      indices.insert(localIndices[i]);
	  }
	  break;
	}
      }

      elInfo = stack.traverseNext(elInfo);
    }
  }

  inline void getImplicitIndices(const FiniteElemSpace* feSpace,
			  AbstractFunction<double, WorldVector<double> >* signedDistFct,
			  std::vector<DegreeOfFreedom> &indices)
  {
    const BasisFunction *vecBasisFcts = feSpace->getBasisFcts();
    int nVecBasisFcts = vecBasisFcts->getNumber();

    std::vector<DegreeOfFreedom> vecLocalIndices(nVecBasisFcts);
    
    DOFVector<short int> visited(feSpace, "visited"); visited.set(0);
    WorldVector<double> coord;
    TraverseStack stack;
    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1, traverseFlag);

    while (elInfo) {
      Element *el = elInfo->getElement();
      vecBasisFcts->getLocalIndices(el, feSpace->getAdmin(), vecLocalIndices);

      for (int i = 0; i < nVecBasisFcts; i++) {
	if ( visited[vecLocalIndices[i]] == 0 ) {
	  elInfo->coordToWorld(*(vecBasisFcts->getCoords(i)), coord);
	  if ( (*signedDistFct)(coord) <= DBL_TOL ) { // eventuell nur Koordinaten an Gitterknoten bekommen!
	    indices.push_back(vecLocalIndices[i]);
	  }
	  visited[vecLocalIndices[i]] = 1;
	}
      }
      elInfo = stack.traverseNext(elInfo); 
    }
  }
  
  inline void getImplicitIndices(const FiniteElemSpace* feSpace,
			  DOFVector<double>* signedDistDOF,
			  std::vector<DegreeOfFreedom> &indices)
  {
   
    const BasisFunction *vecBasisFcts = feSpace->getBasisFcts();
    int nVecBasisFcts = vecBasisFcts->getNumber();

    std::vector<DegreeOfFreedom> vecLocalIndices(nVecBasisFcts);
    ElementVector signedDistLocalCoeffs(nVecBasisFcts);

    DimVec<double> *baryCoords = NULL;
    DOFVector<short int> visited(feSpace, "visited"); visited.set(0);
    TraverseStack stack;
    Flag traverseFlag = Mesh::CALL_LEAF_EL;
    ElInfo *elInfo = stack.traverseFirst(feSpace->getMesh(), -1, traverseFlag);

    while (elInfo) {
      Element *el = elInfo->getElement();

      vecBasisFcts->getLocalIndices(el, feSpace->getAdmin(), vecLocalIndices);
      signedDistDOF->getLocalVector(el, signedDistLocalCoeffs);

      for (int i = 0; i < nVecBasisFcts; i++) {
	if ( visited[vecLocalIndices[i]] == 0 ) {
	  baryCoords = vecBasisFcts->getCoords(i);

	  double dist = vecBasisFcts->evalUh(*baryCoords, signedDistLocalCoeffs);
	  if (dist <= DBL_TOL) {
	    indices.push_back(vecLocalIndices[i]);
	  }
	  visited[vecLocalIndices[i]] = 1;
	}
      }
      elInfo = stack.traverseNext(elInfo); 
    }
  }
  
  /**
   * get al list of pairs that describes wich DegreeOfFreedom are assiciated in the given mesh
   * with given feSpace. Uses the function getBoundaryIndices().
   * The 'nr' and 'meshIndicator' describe one boundary of the mesh and to each DegreeOfFreedom
   * the 'periodicMap' associates a second DegreeOfFreedom.
   **/
//   inline void getPeriodicAssociation(const FiniteElemSpace* feSpace,
// 			      BoundaryType nr,
// 			      AbstractFunction<bool, WorldVector<double> >* meshIndicator,
// 			      AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap,
// 			      std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &association)
//   {
//     std::set<DegreeOfFreedom> indices;
//     details::getBoundaryIndices(feSpace, nr, meshIndicator, indices);
// 
//     association.clear();
// 
//     DOFVector<WorldVector<double> > coords(feSpace, "coords");
//     feSpace->getMesh()->getDofIndexCoords(coords);
//     std::set<DegreeOfFreedom>::iterator it;
//     for (it = indices.begin(); it != indices.end(); it++)
//     {
//       DegreeOfFreedom idx2;
//       WorldVector<double> p;
//       p = (*periodicMap)(coords[*it]);
//       TEST_EXIT(coords.getDofIdxAtPoint(p,idx2))("periodic association not found!\n");
// 
//       association.push_back(std::make_pair(min(*it, idx2), max(*it, idx2)));
//     }
//   }

  /**
   * get al list of pairs that describes wich DegreeOfFreedom are assiciated in the given mesh
   * with given feSpace. Uses the function getBoundaryIndices().
   * The 'meshIndicator' describes one boundary of the mesh and to each DegreeOfFreedom
   * the 'periodicMap' associates a second DegreeOfFreedom.
   **/
//     inline void getPeriodicAssociation(const FiniteElemSpace* feSpace,
// 			      AbstractFunction<bool, WorldVector<double> >* meshIndicator,
// 			      AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap,
// 			      std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &association)
//   {
//     std::set<DegreeOfFreedom> indices;
//     details::getBoundaryIndices(feSpace, meshIndicator, indices);
//     MSG_DBG("Nr of indices: %d\n", indices.size());
// 
//     association.clear();
// 
//     DOFVector<WorldVector<double> > coords(feSpace, "coords");
//     feSpace->getMesh()->getDofIndexCoords(coords);
//     std::set<DegreeOfFreedom>::iterator it;
//     for (it = indices.begin(); it != indices.end(); it++)
//     {
//       DegreeOfFreedom idx2;
//       WorldVector<double> p;
//       p = (*periodicMap)(coords[*it]);
//       TEST_EXIT(coords.getDofIdxAtPoint(p,idx2))("periodic association not found!\n");
// 
//       association.push_back(std::make_pair(min(*it, idx2), max(*it, idx2));
//     }
//     
//     MSG_DBG("Nr of associations: %d\n", association.size());
//   }
  
  inline void getPeriodicAssociation(const FiniteElemSpace* feSpace,
				     std::set<DegreeOfFreedom> indices,
				     AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap,
				     std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &association)
  {
    association.clear();

    DOFVector<WorldVector<double> > coords(feSpace, "coords");
    feSpace->getMesh()->getDofIndexCoords(coords);
    std::set<DegreeOfFreedom>::iterator it;
    
    DegreeOfFreedom idx_;
    WorldVector<double> p;
    for (it = indices.begin(); it != indices.end(); it++)
    {
      p = (*periodicMap)(coords[*it]);
      TEST_EXIT(coords.getDofIdxAtPoint(p,idx_))("periodic association not found!\n");
      association.push_back(std::make_pair(std::min(*it, idx_), std::max(*it, idx_)));
    }
    
#if DEBUG != 0
    MSG("Nr of associations: %d\n", association.size());
#endif
  }
  
  
  inline void getDOFValues(const FiniteElemSpace* feSpace, AbstractFunction<double, WorldVector<double> >* values, std::vector<DegreeOfFreedom> &indices, std::vector<double> &dofValues)
  {
    WorldVector<double> x;
    for (size_t i = 0; i < indices.size(); i++) {
      if (feSpace->getMesh()->getDofIndexCoords(indices[i], feSpace, x))
	dofValues.push_back((*values)(x));
    }
  }
  
  inline void getDOFValues(const FiniteElemSpace* feSpace, DOFVector<double>* values, std::vector<DegreeOfFreedom> &indices, std::vector<double> &dofValues)
  {
    for (size_t i = 0; i < indices.size(); i++) {
      dofValues.push_back((*values)[indices[i]]);
    }
  }
} // end namespace



/**
 * data structure that holds all possible variation of description of coords and value
 * of (singular/implicit) dirichlet boundary condition.
 **/
struct DirichletBcData {

  // pos = WorldVector
  DirichletBcData(int i_, int j_, WorldVector<double> pos_, double val_)
  : row(i_), col(j_), pos(pos_), val0(val_), val1(NULL), val2(NULL),
    posType(0), valueType(0),
    idx(0), SignedDistFct(NULL), SignedDistDOF(NULL),
    boundary_nr(0), meshIndicator(NULL) {}

  DirichletBcData(int i_, int j_, WorldVector<double> pos_, DOFVector<double> &val_)
  : row(i_), col(j_), pos(pos_), val1(&val_), val0(0.0), val2(NULL),
    posType(0), valueType(1),
    idx(0), SignedDistFct(NULL), SignedDistDOF(NULL),
    boundary_nr(0), meshIndicator(NULL)  {}

  DirichletBcData(int i_, int j_, WorldVector<double> pos_, AbstractFunction<double, WorldVector<double> > &val_)
  : row(i_), col(j_), pos(pos_), val0(0.0), val1(NULL), val2(&val_),
    posType(0), valueType(2),
    idx(0), SignedDistFct(NULL), SignedDistDOF(NULL), 
    boundary_nr(0), meshIndicator(NULL)  {}

  // pos = idx
  DirichletBcData(int i_, int j_, DegreeOfFreedom idx_, double val_)
  : row(i_), col(j_), idx(idx_), val0(val_), val1(NULL), val2(NULL),
    posType(1), valueType(0),
    SignedDistFct(NULL), SignedDistDOF(NULL), 
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, DegreeOfFreedom idx_, DOFVector<double> &val_)
  : row(i_), col(j_), idx(idx_), val0(0.0), val1(&val_), val2(NULL),
    posType(1), valueType(1),
    SignedDistFct(NULL), SignedDistDOF(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, DegreeOfFreedom idx_, AbstractFunction<double, WorldVector<double> > &val_)
  : row(i_), col(j_), idx(idx_), val0(0.0), val1(NULL), val2(&val_),
    posType(1), valueType(2),
    SignedDistFct(NULL), SignedDistDOF(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  // pos = SignedDistFct
  DirichletBcData(int i_, int j_, AbstractFunction<double, WorldVector<double> > *SignedDistFct_, double val_)
  : row(i_), col(j_), SignedDistFct(SignedDistFct_), val0(val_), val1(NULL), val2(NULL),
    posType(2), valueType(0),
    idx(0), SignedDistDOF(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, AbstractFunction<double, WorldVector<double> > *SignedDistFct_, DOFVector<double> &val_)
  : row(i_), col(j_), SignedDistFct(SignedDistFct_), val0(0.0), val1(&val_), val2(NULL),
    posType(2), valueType(1),
    idx(0), SignedDistDOF(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, AbstractFunction<double, WorldVector<double> > *SignedDistFct_, AbstractFunction<double, WorldVector<double> > &val_)
  : row(i_), col(j_), SignedDistFct(SignedDistFct_), val0(0.0), val1(NULL), val2(&val_),
    posType(2), valueType(2),
    idx(0), SignedDistDOF(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  // pos = SignedDistDOF
  DirichletBcData(int i_, int j_, DOFVector<double> *SignedDistDOF_, double val_)
  : row(i_), col(j_), SignedDistDOF(SignedDistDOF_), val0(val_), val1(NULL), val2(NULL),
    posType(3), valueType(0),
    idx(0), SignedDistFct(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, DOFVector<double> *SignedDistDOF_, DOFVector<double> &val_)
  : row(i_), col(j_), SignedDistDOF(SignedDistDOF_), val0(0.0), val1(&val_), val2(NULL),
    posType(3), valueType(1),
    idx(0), SignedDistFct(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, DOFVector<double> *SignedDistDOF_, AbstractFunction<double, WorldVector<double> > &val_)
  : row(i_), col(j_), SignedDistDOF(SignedDistDOF_), val0(0.0), val1(NULL), val2(&val_),
    posType(3), valueType(2),
    idx(0), SignedDistFct(NULL),
    boundary_nr(0), meshIndicator(NULL)  { pos.set(0.0); }
    
  // pos = meshindicator + boundary_nr
  DirichletBcData(int i_, int j_, BoundaryType nr, AbstractFunction<bool, WorldVector<double> > *meshIndicator_, double val_)
  : row(i_), col(j_), SignedDistDOF(NULL), val0(val_), val1(NULL), val2(NULL),
    posType(4), valueType(0),
    idx(0), SignedDistFct(NULL),
    boundary_nr(nr), meshIndicator(meshIndicator_)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, BoundaryType nr, AbstractFunction<bool, WorldVector<double> > *meshIndicator_, DOFVector<double> &val_)
  : row(i_), col(j_), SignedDistDOF(NULL), val0(0.0), val1(&val_), val2(NULL), 
    posType(4), valueType(1),
    idx(0), SignedDistFct(NULL),
    boundary_nr(nr), meshIndicator(meshIndicator_)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, BoundaryType nr, AbstractFunction<bool, WorldVector<double> > *meshIndicator_, AbstractFunction<double, WorldVector<double> > &val_)
  : row(i_), col(j_), SignedDistDOF(NULL), val0(0.0), val1(NULL), val2(&val_),
    posType(4), valueType(2),
    idx(0), SignedDistFct(NULL),
    boundary_nr(nr), meshIndicator(meshIndicator_)  { pos.set(0.0); }
    
  // pos = meshindicator
  DirichletBcData(int i_, int j_, AbstractFunction<bool, WorldVector<double> > *meshIndicator_, double val_)
  : row(i_), col(j_), SignedDistDOF(NULL), val0(val_), val1(NULL), val2(NULL),
    posType(4), valueType(0),
    idx(0), SignedDistFct(NULL),
    boundary_nr(0), meshIndicator(meshIndicator_)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, AbstractFunction<bool, WorldVector<double> > *meshIndicator_, DOFVector<double> &val_)
  : row(i_), col(j_), SignedDistDOF(NULL), val0(0.0), val1(&val_), val2(NULL),
    posType(4), valueType(1),
    idx(0), SignedDistFct(NULL),
    boundary_nr(0), meshIndicator(meshIndicator_)  { pos.set(0.0); }

  DirichletBcData(int i_, int j_, AbstractFunction<bool, WorldVector<double> > *meshIndicator_, AbstractFunction<double, WorldVector<double> > &val_)
  : row(i_), col(j_), SignedDistDOF(NULL), val0(0.0), val1(NULL), val2(&val_),
    posType(4), valueType(2),
    idx(0), SignedDistFct(NULL),
    boundary_nr(0), meshIndicator(meshIndicator_)  { pos.set(0.0); }
    
  void addToList(const FiniteElemSpace *feSpace, std::vector<SingularDirichletBC> &list)
  {
    std::vector<DegreeOfFreedom> indices_;
    std::vector<double> values_;
    DegreeOfFreedom idx_ = 0;

    switch (posType) {
      case 0:
	if (val1 == NULL && SignedDistDOF == NULL) {
	  val1 = new DOFVector<double>(feSpace, "vel1");
	}
	if (val1 != NULL) {
	  if (!val1->getDofIdxAtPoint(pos, idx_))
	    idx_ = 0;
	} else if (SignedDistDOF != NULL) {
	  if (!SignedDistDOF->getDofIdxAtPoint(pos, idx_))
	    idx_ = 0;
	}
	indices_.push_back(idx_);
	break;
	
      case 1:
	indices_.push_back(idx);
	break;
	
      case 2:
	details::getImplicitIndices(feSpace, SignedDistFct, indices_);
	break;
	
      case 3:
	details::getImplicitIndices(feSpace, SignedDistDOF, indices_);
	break;
	
      case 4:
	std::set<DegreeOfFreedom> indices__;
	if (boundary_nr == 0)
	  details:: getBoundaryIndices(feSpace, meshIndicator, indices__);
	else
	  details:: getBoundaryIndices(feSpace, boundary_nr, meshIndicator, indices__);
	std::set<DegreeOfFreedom>::iterator it;
	for (it = indices__.begin(); it != indices__.end(); it++)
	  indices_.push_back(*it);
	break;
    }

    switch (valueType) {
      case 0:
	values_.resize(indices_.size(), val0);
	break;
	
      case 1:
	details::getDOFValues(feSpace, val1, indices_, values_);
	break;
	
      case 2:
	details::getDOFValues(feSpace, val2, indices_, values_);
	break;
    }

    DOFVector<WorldVector<double> > coords(feSpace, "coords");
    feSpace->getMesh()->getDofIndexCoords(coords);
    

    for (size_t i = 0; i < indices_.size(); i++) {
      list.push_back(SingularDirichletBC(row, col, indices_[i], values_[i]));
    }
    
#if DEBUG != 0
    MSG("Dirichle BC at %d DOFs added.\n",indices_.size());
#endif
  }


  int row,col;

private:
  int posType; 		// 0..WorldVector, 1..DegreeOfFreedom, 2..SignedDistFct, 3..signedDistDOF, 4..meshIndicator
  int valueType; 	// 0..double, 1..DOFVector, 2..AbstractFunction

  // pos
  WorldVector<double> pos;
  DegreeOfFreedom idx;
  AbstractFunction<double, WorldVector<double> > *SignedDistFct;
  DOFVector<double> *SignedDistDOF;

  BoundaryType boundary_nr;
  AbstractFunction<bool, WorldVector<double> >* meshIndicator;
  
  // value
  double val0;
  DOFVector<double> *val1;
  AbstractFunction<double, WorldVector<double> > *val2;
};


struct PeriodicBcData {

  // pos = WorldVector
  PeriodicBcData(int row_, BoundaryType nr_, AbstractFunction<bool, WorldVector<double> >* meshIndicator_,
		 AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap_)
  : row(row_), nr(nr_), meshIndicator(meshIndicator_), periodicMap(periodicMap_) {}

  PeriodicBcData(int row_, AbstractFunction<bool, WorldVector<double> >* meshIndicator_,
		 AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap_)
  : row(row_), nr(0), meshIndicator(meshIndicator_), periodicMap(periodicMap_) {}

  void addToList(const FiniteElemSpace *feSpace, std::vector<ManualPeriodicBC> &list)
  {
    std::set<DegreeOfFreedom> indices;
    if (nr == 0)
      details::getBoundaryIndices(feSpace, meshIndicator, indices);
    else
      details::getBoundaryIndices(feSpace, nr, meshIndicator, indices);
    
    std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > associations;
    details::getPeriodicAssociation(feSpace, indices, periodicMap, associations);
    list.push_back(ManualPeriodicBC(row, associations));
  }
  int row;

protected:


  BoundaryType nr;
  AbstractFunction<bool, WorldVector<double> >* meshIndicator;
  AbstractFunction<WorldVector<double>, WorldVector<double> >* periodicMap;
};

} // end namespace 

#endif // SINGULAR_DIRICHLET_BC_H
