//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.

/** \file VtuReader.hh */

namespace AMDiS {

  namespace VtuReader {
    
    /** \brief
     *
     * The DOF vector must have been created by a corresponding mesh file. The
     * function now reads the corresponding VTU file and
     * copies the values to the correct positions in the DOF vector.
     */
    template<typename T>
    void readValue(std::string filename,
		    Mesh *mesh,
		    DOFVector<T> *dofVector,
		    std::string componentName)
    {
      std::vector<DOFVector<T>*> dofVectors; dofVectors.push_back(dofVector);
      std::vector<std::string> componentNames; componentNames.push_back(componentName);

      readValue(filename, mesh, dofVectors, componentNames);
    }

    template<typename T>
    void readValue(std::string filename,
		    Mesh *mesh,
		    std::vector<DOFVector<T>*> dofVectors,
		    std::vector<std::string> componentNames)
    {
      using namespace pugi;
      using namespace VtuReader::details;

      FUNCNAME("VtuReader::readValue()");

      TEST_EXIT(filename != "")("Filename not specified!\n");
      TEST_EXIT(mesh)("no mesh specified\n");
      TEST_EXIT(dofVectors.size() > 0)("no DOF vectors specified\n");
      TEST_EXIT(componentNames.size() > 0)("no componentName specified\n");

      if(!boost::filesystem::exists(filename))
	throw(std::runtime_error(filename + " does not exist!"));

      xml_document vtu;
      if(!vtu.load_file(filename.c_str()))
	throw(std::runtime_error("Could not load vtu file! Error in xml structure."));

      xml_node VTKFile = vtu.child("VTKFile");
      xml_node UnstructuredGrid = VTKFile.child("UnstructuredGrid");
      xml_node Piece = UnstructuredGrid.child("Piece");
      xml_node PointData = Piece.child("PointData");
      xml_node Points = Piece.child("Points");

      std::string points = Points.child_value("DataArray");
      typedef std::vector<WorldVector<double> > PL;
      PL pointList;
      string2pointList(points, pointList);

      std::vector<std::vector<T> > valueList(dofVectors.size());
      T test;

      for (xml_node DataArray = PointData.child("DataArray"); DataArray; DataArray = DataArray.next_sibling("DataArray")) {
	std::string Name = DataArray.attribute("Name").value();
	for (size_t i = 0; i < componentNames.size(); i++) {
	  if (Name == componentNames[i]) {
	    std::string values = DataArray.last_child().value();
	    int nComponents = -1;
	    if (DataArray.attribute("NumberOfComponents"))
	      nComponents = DataArray.attribute("NumberOfComponents").as_int();
	    if (nComponents != -1 && static_cast<int>(vector_operations::num_rows(test)) > nComponents)
	      throw(std::runtime_error("Can not store values in DOFVector with given value type. Too many components!"));
	    string2valueList(values, valueList[i], vector_operations::num_rows(test), nComponents);
	    break;
	  }
	}
      }
      for (size_t i = 0; i < dofVectors.size(); i++) {
	if (valueList[i].size() == 0)
	  throw(std::runtime_error("no values found for component name '"+componentNames[i]+"'!"));
	if (pointList.size() != valueList[i].size())
	  throw(std::runtime_error("Coordinates and values["+boost::lexical_cast<std::string>(i)+"] do not match!"));
      }

      DOFVector<WorldVector<double> > coords(dofVectors[0]->getFeSpace(), "coords");
      dofVectors[0]->getFeSpace()->getMesh()->getDofIndexCoords(coords);

      experimental::KD_Tree tree(Global::getGeo(WORLD), pointList, 10);
      tree.index->buildIndex();

      DOFIterator<WorldVector<double> > it_coords(&coords, USED_DOFS);
      std::vector<DOFIterator<T>*> it_results;
      for (size_t i = 0; i < dofVectors.size(); i++) {
	it_results.push_back(new DOFIterator<T>(dofVectors[i], USED_DOFS));
	it_results[i]->reset();
      }
      for (it_coords.reset(); !it_coords.end(); ++it_coords) {
	size_t idx = getNearestIndex(tree, *it_coords);

	for (size_t i = 0; i < dofVectors.size(); i++) {
	  if (valueList[i].size() > idx)
	    *(*it_results[i]) = valueList[i][idx];
	  (*it_results[i])++;
	}
      }
      
      
      MSG("VTU file read from: %s\n", filename.c_str());
    }
  } // end namespace VtuReader
  
} // end namespace
