#include "CahnHilliardNavierStokes.h"
#include "Views.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"
#include "POperators.h"

using namespace AMDiS;

CahnHilliardNavierStokes::CahnHilliardNavierStokes(const std::string &name_) :
  super(name_),
  reinit(NULL),
  velocity(NULL),
  useMobility(false),
  useConservationForm(false),
  laplaceType(0),
  nonLinTerm(3),
  doubleWell(1),
  oldTimestep(0.0),
  viscosity(1.0),
  density(1.0),
  c(0.0),
  gamma(1.0),
  eps(0.1),
  minusEps(-0.1),
  epsInv(10.0),
  minusEpsInv(-10.0),
  epsSqr(0.01),
  minusEpsSqr(-0.01),
  sigma(1.0),
  surfaceTension(1.0),
  fileWriter(NULL)
{
  // parameters for CH
  Parameters::get(name + "->use mobility", useMobility); // dt(c) = div(M(c)*grad(c)) + ...
  Parameters::get(name + "->gamma", gamma); // M(c) = gamma * Double-Well
  Parameters::get(name + "->epsilon", eps); // interface width
  
  // type of double well: 0= [0,1], 1= [-1,1]
  Parameters::get(name + "->double-well type", doubleWell); 

  // parameters for navier-stokes
  Initfile::get(name + "->viscosity", viscosity);
  Initfile::get(name + "->density", density);
  
  // type of laplace operator: 	0... div(nu*grad(u)), 
  //							1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 
  
  // type of non-linear term: 	0... u^old*grad(u_i^old), 
  //							1... u*grad(u_i^old), 
  //							2... u^old*grad(u_i)
  //							3... u*grad(u_i^old) + u^old*grad(u_i) - u^old*grad(u_i^old)
  Initfile::get(name + "->non-linear term", nonLinTerm); 

  // Parameters for CH-Coupling
  // 0... u*grad(c), 1... -div(u*c)
  Initfile::get(name + "->use conservation form", useConservationForm);

  // Parameters for NS-Coupling
  // cahn-hiliard-force: sigma*mu*grad(c)
  Initfile::get(name + "->sigma", sigma); // surface tension
  surfaceTension = -sigma*3.0/(2.0*sqrt(2.0)) / eps;

    
  force.set(0.0);
  Initfile::get(name + "->force", force);
  
  // transformation of the parameters
  minusEps = -eps;
  epsInv = 1.0/eps;
  minusEpsInv = -epsInv;
  epsSqr = sqr(eps);
  minusEpsSqr = -epsSqr;
}


CahnHilliardNavierStokes::~CahnHilliardNavierStokes() 
{
  if (reinit != NULL) {
    delete reinit;
    reinit = NULL;
  }
  if (velocity != NULL) {
    delete velocity;
    velocity = NULL;
  }
  delete fileWriter;
  fileWriter = NULL;
}


void CahnHilliardNavierStokes::initData()
{
  dim = getMesh()->getDim();
  // create instance redistancing class
  reinit = new HL_SignedDistTraverse("reinit", dim);
  
  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(2), "velocity");
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace(2)->getMesh(), velocity);

  super::initData();
}


void CahnHilliardNavierStokes::transferInitialSolution(AdaptInfo *adaptInfo)
{
  calcVelocity();
  
  for (int i = 0; i < 2+dow; i++)
    prob->setExactSolution(prob->getSolution()->getDOFVector(i), i);

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);

  // initial parameters for detecting mesh changes
  oldMeshChangeIdx= getMesh()->getChangeIndex();
}


void CahnHilliardNavierStokes::fillOperators()
{ FUNCNAME("CahnHilliardNavierStokes::fillOperators()");
  MSG("CahnHilliardNavierStokes::fillOperators()\n");
  
  // variable order:
  // (c, mu, u0, u1 [, u2], p)
  
  WorldVector<DOFVector<double>* > vel;
  for (size_t i = 0; i < dow; i++)
    vel[i] = prob->getSolution()->getDOFVector(2+i);


  // dt(c) = laplace(mu) - u*grad(c)
  // -----------------------------------
  /// < phi*c/tau , psi >
  Operator *opChMnew = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChMnew->addZeroOrderTerm(new Simple_ZOT);
  prob->addMatrixOperator(*opChMnew, 0, 0, getInvTau()); 
  
  /// < phi*c^old/tau , psi >
  Operator *opChMold = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChMold->addZeroOrderTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(0)));
  prob->addVectorOperator(*opChMold,0, getInvTau());   
  
  /// < phi*grad(mu) , grad(psi) >
  // div(M(c)grad(mu)), with M(c)=gamma/4*(c^2-1)^2
  Operator *opChLM = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
  if (useMobility) {
    if (doubleWell == 0)
      opChLM->addSecondOrderTerm(new VecAtQP_SOT(
	prob->getSolution()->getDOFVector(0),
	new MobilityCH0(gamma)));
    else
      opChLM->addSecondOrderTerm(new VecAtQP_SOT(
	prob->getSolution()->getDOFVector(0),
	new MobilityCH1(gamma)));
  } else
    opChLM->addSecondOrderTerm(new Simple_SOT(gamma));
  prob->addMatrixOperator(*opChLM, 0, 1);
  
  /// < u_old * grad(c) - u_old * grad(c_old), psi >
  Operator *uGradC = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  if (useConservationForm)
    uGradC->addTerm(new WorldVec_FOT(vel, -1.0), GRD_PSI);
  else
    uGradC->addTerm(new WorldVec_FOT(vel), GRD_PHI);
  prob->addMatrixOperator(*uGradC, 0, 0);
  uGradC->setUhOld(prob->getSolution()->getDOFVector(0));
  prob->addVectorOperator(*uGradC, 0);
  
  /// < u * grad(c_old) , psi >
  for (size_t i = 0; i < dow; i++) {
    Operator *uGradC2 = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
    if (useConservationForm)
      uGradC2->addTerm(new VecAndPartialDerivative_FOT(prob->getSolution()->getDOFVector(0), i, -1.0), GRD_PSI);
    else {
      uGradC2->addTerm(new PartialDerivative_ZOT(prob->getSolution()->getDOFVector(0), i));
    }
    prob->addMatrixOperator(*uGradC2, 0, 2+i);
  }
  

  // mu + eps^2*laplace(c) + c - 3*(c_old^2)*c = -2*c_old^3 [+ BC]
  // ----------------------------------------------------------------------
  /// < -3*phi*c*c_old^2 , psi >
  // -3*c_old^2 * c
  Operator *opChMPowImpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChMPowImpl->addZeroOrderTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(0),
    new Pow2Functor(-3.0)));
  if (doubleWell == 0) {
    opChMPowImpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getSolution()->getDOFVector(0),
      new AMDiS::Factor<double>(3.0)));
    opChMPowImpl->addZeroOrderTerm(new Simple_ZOT(-0.5));
  } else {
    opChMPowImpl->addZeroOrderTerm(new Simple_ZOT(1.0));
  }
  prob->addMatrixOperator(*opChMPowImpl, 1, 0);  
  
  /// < -2*phi*c_old^3 , psi >
  // -2*c_old^3 + 3/2*c_old^2
  Operator *opChMPowExpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChMPowExpl->addZeroOrderTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(0),
    new Pow3Functor(-2.0)));
  if (doubleWell == 0) {
    opChMPowExpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getSolution()->getDOFVector(0),
      new Pow2Functor(3.0/2.0)));
  }
  prob->addVectorOperator(*opChMPowExpl, 1);            
  
  /// < -eps^2*phi*grad(c) , grad(psi) >
  Operator *opChL = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChL->addSecondOrderTerm(new Simple_SOT);
  prob->addMatrixOperator(*opChL, 1, 0, &minusEpsSqr);  
  
  /// < phi*mu , psi >
  Operator *opChM = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChM->addZeroOrderTerm(new Simple_ZOT);
  prob->addMatrixOperator(*opChM, 1, 1);              

  // -----------------------------------------------------------------------------------
  for (size_t i = 0; i < dow; ++i) {
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opTime->addTerm(new Simple_ZOT(density));
    prob->addMatrixOperator(*opTime, 2+i, 2+i, getInvTau(), getInvTau());
    /// < (1/tau)*u_i^old , psi >
    Operator *opTimeOld = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opTimeOld->addTerm(new Phase_ZOT(prob->getSolution()->getDOFVector(2+i), density));
    prob->addVectorOperator(*opTimeOld, 2+i, getInvTau(), getInvTau());
 
    /// < u^old*grad(u_i^old) , psi >
    if (nonLinTerm == 0 || nonLinTerm == 3) {
	  Operator *opUGradU0 = new Operator(getFeSpace(2+i), getFeSpace(2+i));
	  opUGradU0->addTerm(new WorldVec_FOT(vel, (nonLinTerm == 0 ? -1.0 : 1.0)*density), 
						 GRD_PHI);
	  opUGradU0->setUhOld(prob->getSolution()->getDOFVector(2+i));
      prob->addVectorOperator(*opUGradU0, 2+i);
    }
    if (nonLinTerm == 1 || nonLinTerm == 3) {
      /// < u*grad(u_i^old) , psi >
      for (size_t j = 0; j < dow; ++j) {
        Operator *opUGradU1 = new Operator(getFeSpace(2+i),getFeSpace(2+i));
        opUGradU1->addTerm(new PartialDerivative_ZOT(
          prob->getSolution()->getDOFVector(2+i), j, density));
        prob->addMatrixOperator(*opUGradU1, 2+i, 2+j);
      }
    } 
    if (nonLinTerm == 2 || nonLinTerm == 3) {
      /// < u^old*grad(u_i) , psi >
      for(size_t j = 0; j < dow; ++j) {
        Operator *opUGradU2 = new Operator(getFeSpace(2+i),getFeSpace(2+i));
        opUGradU2->addTerm(new VecAndPartialDerivative_FOT(
          prob->getSolution()->getDOFVector(2+j), j, density), GRD_PHI);
        prob->addMatrixOperator(*opUGradU2, 2+i, 2+i);
      }
    }

//     for (size_t j = 0; j < dow; ++j) {
//       Operator *opNull = new Operator(getFeSpace(2+i), getFeSpace(2+j));
//       opNull->addTerm(new Simple_ZOT(0.0));
//       prob->addMatrixOperator(*opNull, 2+i, 2+j);
//     }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity)
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(2+i),getFeSpace(2+dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, 2+i, 2+dow);
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(2+i), getFeSpace(2+i));
      opForce->addTerm(new Simple_ZOT(force[i]));
      prob->addVectorOperator(*opForce, 2+i);
    }
    
    // forces by Cahn-Hilliard terms
    // -----------------------------
    /// < mu_old * grad(c) , psi >
    Operator *opMuGradC = new Operator(getFeSpace(2+i),getFeSpace(0));
    opMuGradC->addTerm(new VecAndPartialDerivative_FOT(
      prob->getSolution()->getDOFVector(1), i), GRD_PHI);
    prob->addMatrixOperator(*opMuGradC, 2+i, 0, &surfaceTension, &surfaceTension);
    
	/// < mu * grad(c_old) , psi >
    Operator *opMuGradC2 = new Operator(getFeSpace(2+i),getFeSpace(1));
    opMuGradC2->addTerm(new PartialDerivative_ZOT(
      prob->getSolution()->getDOFVector(0), i));
    prob->addMatrixOperator(*opMuGradC2, 2+i, 1, &surfaceTension, &surfaceTension);
	
	/// < mu_old * grad(c_old) , psi >
    opMuGradC2->setUhOld(prob->getSolution()->getDOFVector(1));
    prob->addVectorOperator(*opMuGradC2, 2+i, &surfaceTension, &surfaceTension);
    
  }

  /// div(u) = 0
  for (size_t i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(2+dow),getFeSpace(2+i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, 2+dow, 2+i);
  }
}


void CahnHilliardNavierStokes::addLaplaceTerm(int i)
{
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (laplaceType == 1) {
    for (size_t j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(2+i), getFeSpace(2+j));
      opLaplaceUi1->addTerm(new MatrixIJ_SOT(1-i, 1-j, viscosity));
      prob->addMatrixOperator(*opLaplaceUi1, 2+i, 2+j);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opLaplaceUi->addTerm(new Simple_SOT(viscosity));
  prob->addMatrixOperator(*opLaplaceUi, 2+i, 2+i);
}

void CahnHilliardNavierStokes::closeTimestep(AdaptInfo *adaptInfo)
{
  calcVelocity();
  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);
}
