#include "NavierStokes_TH_MultiPhase.h"

using namespace std;
using namespace AMDiS;

NavierStokes_TH_MultiPhase::NavierStokes_TH_MultiPhase(const std::string &name_) :
  super(name_),
  multiPhase(NULL),
  densityPhase(NULL),
  viscosityPhase(NULL),
  viscosity1(1.0),
  viscosity2(1.0),
  density1(1.0),
  density2(1.0)
{
  Initfile::get(name + "->viscosity1", viscosity1); // viscosity of fluid 1
  Initfile::get(name + "->viscosity2", viscosity2); // viscosity of fluid 2
  Initfile::get(name + "->density1", density1); // density of fluid 1
  Initfile::get(name + "->density2", density2); // density of fluid 2

  if (viscosity1 <= 0.0)
    viscosity1 = viscosity;
};


NavierStokes_TH_MultiPhase::~NavierStokes_TH_MultiPhase()
{ FUNCNAME("NavierStokes_TH_MultiPhase::~NavierStokes_TH_MultiPhase()");

  //delete multiPhase;
  delete densityPhase;
  delete viscosityPhase;
};


void NavierStokes_TH_MultiPhase::initData()
{ FUNCNAME("NavierStokes_TH_MultiPhase::initTimeInterface()");

 // multiPhase = new DOFVector<double>(prob->getFeSpace(0), "multiPhase");
  densityPhase = new DOFVector<double>(prob->getFeSpace(0), "densityPhase");
  viscosityPhase = new DOFVector<double>(prob->getFeSpace(0), "viscosityPhase");

//  multiPhase->set(1.0);
  densityPhase->set(density1);
  viscosityPhase->set(viscosity1);

  super::initData();
};

void NavierStokes_TH_MultiPhase::initTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TH_MultiPhase::initTimestep()");

  super::initTimestep(adaptInfo);
  LinearInterpolation1 dLI(density1, density2);
  LinearInterpolation1 vLI(viscosity1, viscosity2);
  transformDOF(multiPhase, densityPhase, &dLI);
  transformDOF(multiPhase, viscosityPhase, &vLI);
};


void NavierStokes_TH_MultiPhase::fillOperators()
{ FUNCNAME("NavierStokes_TH_MultiPhase::fillOperators()");
  WorldVector<DOFVector<double>* > vel;
  for (unsigned i = 0; i < dow; i++){
    vel[i]= prob->getSolution()->getDOFVector(i);
  }
  
  // fill operators for prob
  for (unsigned i = 0; i < dow; ++i) {
    /// < (1/tau)*rho*u'_i , psi >
    Operator *opTime = new Operator(getFeSpace(i), getFeSpace(i));
    if (density1==density2)
      opTime->addTerm(new Simple_ZOT());  
    else
      opTime->addTerm(new VecAtQP_ZOT(densityPhase, NULL));
    prob->addMatrixOperator(*opTime, i, i, getInvTau(), getInvTau());
    /// < (1/tau)*rho*u_i^old , psi >
    Operator *opTimeOld = new Operator(getFeSpace(i), getFeSpace(i));
    if (density1==density2)
      opTimeOld->addTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(i), NULL));  
    else
      opTimeOld->addTerm(new Vec2AtQP_ZOT(densityPhase, prob->getSolution()->getDOFVector(i), NULL));
    prob->addVectorOperator(*opTimeOld, i, getInvTau(), getInvTau());
 
    /// < u^old*grad(u_i^old) , psi >
    Operator *opUGradU0 = new Operator(getFeSpace(i),getFeSpace(i));
    if (density1==density2)
        opUGradU0->addTerm(new WorldVec_FOT(vel, -1.0), GRD_PHI);
    else
        opUGradU0->addTerm(new WorldVecPhase_FOT(densityPhase, vel, -1.0), GRD_PHI);
    opUGradU0->setUhOld(prob->getSolution()->getDOFVector(i));
    if (nonLinTerm == 0) {
      prob->addVectorOperator(*opUGradU0, i);
    } else {
      prob->addVectorOperator(*opUGradU0, i, &theta1, &theta1);
    }

    if (nonLinTerm == 1) {
      /// < u'*grad(u_i^old) , psi >
      for (unsigned j = 0; j < dow; ++j) {
        Operator *opUGradU1 = new Operator(getFeSpace(i),getFeSpace(i));
	if (density1==density2)
	  opUGradU1->addTerm(new PartialDerivative_ZOT(prob->getSolution()->getDOFVector(i), j));
	else
	  opUGradU1->addTerm(new VecAndPartialDerivative_ZOT(     densityPhase,    prob->getSolution()->getDOFVector(i), j));
        prob->addMatrixOperator(*opUGradU1, i, j, &theta, &theta);
      }
    } else if (nonLinTerm == 2) {
      /// < u^old*grad(u'_i) , psi >
      for(unsigned j = 0; j < dow; ++j) {
        Operator *opUGradU2 = new Operator(getFeSpace(i),getFeSpace(i));
        opUGradU2->addTerm(new Vec2ProductPartial_FOT(   densityPhase,    prob->getSolution()->getDOFVector(j), j), GRD_PHI);
        prob->addMatrixOperator(*opUGradU2, i, i, &theta, &theta);
      }
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i),getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, i, dow);
    
    /// external force, i.e. gravitational force
    if (force[i]*force[i] > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      opForce->addZeroOrderTerm(new Phase_ZOT(densityPhase, force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (unsigned i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow),getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, dow, i);
  }
};


void NavierStokes_TH_MultiPhase::addLaplaceTerm(int i)
{ FUNCNAME("NavierStokes_TH_MultiPhase::addLaplaceTerm()");

  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (viscosity1!=viscosity2) {
    for (unsigned j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
      opLaplaceUi1->addTerm(new VecAtQP_IJ_SOT(  viscosityPhase, NULL, j, i));
      opLaplaceUi1->setUhOld(prob->getSolution()->getDOFVector(j));
      prob->addMatrixOperator(*opLaplaceUi1, i, j, &theta, &theta);
      prob->addVectorOperator(*opLaplaceUi1, i, &minusTheta1, &minusTheta1);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
  opLaplaceUi->addTerm(new VecAtQP_SOT(viscosityPhase, NULL));
  opLaplaceUi->setUhOld(getSolution()->getDOFVector(i));
  prob->addMatrixOperator(*opLaplaceUi, i, i, &theta, &theta);  
  prob->addVectorOperator(*opLaplaceUi, i, &minusTheta1, &minusTheta1);
};
