#include "AMDiS.h"
#include "NavierStokesPhase_TaylorHood.h"
#include "navierStokes.h"
// #include "time/ExtendedRosenbrockAdaptInstationary.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"

#include "boost/date_time/posix_time/posix_time.hpp"

using namespace AMDiS;
using namespace boost::posix_time;

class NS_Channel : public NavierStokesPhase_TaylorHood
{
public:
  typedef NavierStokesPhase_TaylorHood super;
  
public:
  NS_Channel(std::string name_) : super(name_), phaseField(NULL), inflowBC(NULL) {}
  
  ~NS_Channel()
  {
    if (phaseField != NULL) {
      delete phaseField;
      phaseField = NULL;
    }
    if (inflowBC != NULL) {
      delete inflowBC;
      inflowBC = NULL;
    }
  }
  
  void initData()
  { FUNCNAME("NS_Channel::initData()");
  
    super::initData();
  
    phaseField = new DOFVector<double>(getFeSpace(0), "phaseField"); 
    phaseField->set(1.0);
    super::setPhase(phaseField);   
    
    double H = 4.1;
    double Um = 1.5;
    Parameters::get("mesh->H",H);
    Parameters::get("ns->Um",Um);
    inflowBC = new InflowBC(H,Um);
  }

  void solveInitialProblem(AdaptInfo *adaptInfo)
  { FUNCNAME("NS_Channel::solveInitialProblem()");
  
    super::solveInitialProblem(adaptInfo);

    double radius = 0.1;
    WorldVector<double> center;
    Parameters::get("obstacle->center",center);
    Parameters::get("obstacle->radius",radius);
    obstacle = new Circle(radius, center);

    refFunction = new SignedDistRefinement(getMesh());
    refinement = new RefinementLevelCoords2(getFeSpace(), refFunction, obstacle);

    // initial refinement
    refinement->refine(10);    
    phaseField->interpol(new SignedDistFctToPhaseField(getEpsilon(), obstacle, -3.0));
    
    if (inflowBC)
      inflowBC->setTimePtr(adaptInfo->getTimePtr());
  }
  
protected:
  
  void fillBoundaryConditions()
  { FUNCNAME("NS_Channel::fillBoundaryConditions()");
    
    WorldVector<double> zeroVec; zeroVec.set(0.0);
    AbstractFunction<WorldVector<double>, WorldVector<double> > *zeroVecFct = new AMDiS::Const<WorldVector<double>, WorldVector<double> >(zeroVec);

    super::setBcFct(zeroVecFct);    
    super::fillBoundaryConditions();
    
    
    // +------ 5 ------+
    // |               |
    // 2   # <--1      3
    // |               |
    // +------ 4 ------+
    
    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    
    /// at rigid wall: no-slip boundary condition
    for (size_t i = 0; i < dow; i++) {
      getProblem(0)->addDirichletBC(4, i, i, zero);
      getProblem(0)->addDirichletBC(5, i, i, zero);
    }
    /// dirichlet bc for pressure at one DOF
//     getProblem(0)->addSingularDirichletBC(0, dow, dow, *zero);


    /// at left wall: prescribed velocity    
    if (inflowBC) {
      getProblem(0)->addDirichletBC(2, 0, 0, inflowBC);
      getProblem(0)->addDirichletBC(2, 1, 1, zero);
    }
  }
  
private:
  
  DOFVector<double>* phaseField;
  AbstractFunction<double, WorldVector<double> >* obstacle;
  
  SignedDistRefinement* refFunction;
  RefinementLevelCoords2* refinement;
  
  InflowBC* inflowBC;
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);

  NS_Channel nsProb("ns");
  nsProb.initialize(INIT_ALL | INIT_EXACT_SOLUTION);

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", nsProb.getNumComponents());

  // adaption loop - solve ch-prob and ns-prob
  AdaptInstationary adaptInstat("adapt", nsProb, adaptInfo, nsProb, adaptInfo);
//   ExtendedRosenbrockAdaptInstationary<NS_Channel> adaptInstat("adapt", nsProb, adaptInfo, nsProb, adaptInfo);
  
  // scale Mesh
  WorldVector<double> scale; scale[0] = 25.0; scale[1] = 4.1;
  Helpers::scaleMesh(nsProb.getMesh(), scale);

  ptime start_time = microsec_clock::local_time();
  nsProb.initTimeInterface();
  int error_code = adaptInstat.adapt(); 
  time_duration td = microsec_clock::local_time()-start_time;

  MSG("elapsed time= %d sec\n", td.total_seconds());

  AMDiS::finalize();

  return error_code;
};
